package wacc

import scala.collection.mutable.LinkedHashMap
import scala.collection.mutable.ListBuffer

object asmGenerator {
  import microWacc._
  import assemblyIR._
  import wacc.types._

  val RAX = Register(RegSize.R64, RegName.AX)
  val EAX = Register(RegSize.E32, RegName.AX)
  val RSP = Register(RegSize.R64, RegName.SP)
  val ESP = Register(RegSize.E32, RegName.SP)
  val EDX = Register(RegSize.E32, RegName.DX)
  val RDI = Register(RegSize.R64, RegName.DI)
  val RIP = Register(RegSize.R64, RegName.IP)
  val RBP = Register(RegSize.R64, RegName.BP)
  val RSI = Register(RegSize.R64, RegName.SI)
  val RDX = Register(RegSize.R64, RegName.DX)
  val RCX = Register(RegSize.R64, RegName.CX)
  val R8 = Register(RegSize.R64, RegName.Reg8)
  val R9 = Register(RegSize.R64, RegName.Reg9)

  val _8_BIT_MASK = 0xff

  object labelGenerator {
    var labelVal = -1
    def getLabel(): String = {
      labelVal += 1
      s".L$labelVal"
    }
    def getLabel(target: CallTarget): String = target match {
      case Ident(v, _)   => s"wacc_$v"
      case Builtin(name) => s"_$name"
    }
  }

  def generateAsm(microProg: Program): List[AsmLine] = {
    given stack: LinkedHashMap[Ident, Int] = LinkedHashMap[Ident, Int]()
    given strings: ListBuffer[String] = ListBuffer[String]()
    val Program(funcs, main) = microProg

    val progAsm =
      LabelDef("main") ::
        funcPrologue() ++
        alignStack() ++
        main.flatMap(generateStmt) ++
        List(Move(RAX, ImmediateVal(0))) ++
        funcEpilogue() ++
        generateBuiltInFuncs()

    val strDirs = strings.toList.zipWithIndex.flatMap { case (str, i) =>
      List(Directive.Int(str.size), LabelDef(s".L.str$i"), Directive.Asciz(str.replace("\"", "\\\"")))
    }

    List(Directive.IntelSyntax, Directive.Global("main"), Directive.RoData) ++
      strDirs ++
      List(Directive.Text) ++
      progAsm
  }

  def wrapFunc(labelName: String, funcBody: List[AsmLine])(using
      stack: LinkedHashMap[Ident, Int],
      strings: ListBuffer[String]
  ): List[AsmLine] = {
    LabelDef(labelName) ::
      funcPrologue() ++
      funcBody ++
      funcEpilogue()
  }

  def generateBuiltInFuncs()(using
      stack: LinkedHashMap[Ident, Int],
      strings: ListBuffer[String]
  ): List[AsmLine] = {
    wrapFunc(
      labelGenerator.getLabel(Builtin.Exit),
      alignStack() ++
        List(assemblyIR.Call(CLibFunc.Exit))
    ) ++
      wrapFunc(
        labelGenerator.getLabel(Builtin.Printf),
        alignStack() ++
          List(
            assemblyIR.Call(CLibFunc.PrintF),
            Move(RDI, ImmediateVal(0)),
            assemblyIR.Call(CLibFunc.Fflush)
          )
      ) ++
      wrapFunc(
        labelGenerator.getLabel(Builtin.Malloc),
        alignStack() ++
          List()
      ) ++
      wrapFunc(labelGenerator.getLabel(Builtin.Free), List()) ++
      wrapFunc(
        labelGenerator.getLabel(Builtin.Read),
        alignStack() ++
          List(
            Push(RSI),
            Load(RSI, MemLocation(RSP)),
            assemblyIR.Call(CLibFunc.Scanf),
            Pop(RAX)
          )
      )
  }

  def generateStmt(
      stmt: Stmt
  )(using stack: LinkedHashMap[Ident, Int], strings: ListBuffer[String]): List[AsmLine] =
    stmt match {
      case Assign(lhs, rhs) =>
        var dest: IndexAddress =
          IndexAddress(RSP, 0) // gets overrwitten
        (lhs match {
          case ident: Ident =>
            if (!stack.contains(ident)) {
              stack += (ident -> (stack.size + 1))
              dest = accessVar(ident)
              List(Subtract(RSP, ImmediateVal(8)))
            } else {
              dest = accessVar(ident)
              List()
            }
          // TODO lhs = arrayElem
          case _ =>
            // dest = ???
            List()
        }) ++
          evalExprOntoStack(rhs) ++
          List(Pop(dest))
      case If(cond, thenBranch, elseBranch) => {
        val elseLabel = labelGenerator.getLabel()
        val endLabel = labelGenerator.getLabel()
        evalExprOntoStack(cond) ++
          List(
            Compare(MemLocation(RSP, SizeDir.Word), ImmediateVal(0)),
            Add(RSP, ImmediateVal(8)),
            Jump(LabelArg(elseLabel), Cond.Equal)
          ) ++
          thenBranch.flatMap(generateStmt) ++
          List(Jump(LabelArg(endLabel)), LabelDef(elseLabel)) ++
          elseBranch.flatMap(generateStmt) ++
          List(LabelDef(endLabel))
      }
      case While(cond, body) => {
        val startLabel = labelGenerator.getLabel()
        val endLabel = labelGenerator.getLabel()
        List(LabelDef(startLabel)) ++
          evalExprOntoStack(cond) ++
          List(
            Compare(MemLocation(RSP, SizeDir.Word), ImmediateVal(0)),
            Add(RSP, ImmediateVal(8)),
            Jump(LabelArg(endLabel), Cond.Equal)
          ) ++
          body.flatMap(generateStmt) ++
          List(Jump(LabelArg(startLabel)), LabelDef(endLabel))
      }
      case microWacc.Return(expr) =>
        evalExprOntoStack(expr) ++
          List(Pop(RAX), assemblyIR.Return())
      case call: microWacc.Call => generateCall(call)
    }

  def evalExprOntoStack(expr: Expr)(using
      stack: LinkedHashMap[Ident, Int],
      strings: ListBuffer[String]
  ): List[AsmLine] = {
    expr match {
      case IntLiter(v) =>
        List(Push(ImmediateVal(v)))
      case CharLiter(v) =>
        List(Push(ImmediateVal(v.toInt)))
      case ident: Ident =>
        List(Push(accessVar(ident)))
      case ArrayLiter(elems) =>
        expr.ty match {
          case KnownType.String =>
            strings += elems.map {
              case CharLiter(v) => v
              case _            => ""
            }.mkString
            List(
              Load(
                RAX,
                IndexAddress(
                  RIP,
                  LabelArg(s".L.str${strings.size - 1}")
                )
              ),
              Push(RAX)
            )
          // TODO other array types
          case _ => List()
        }
      case BoolLiter(v)              => List(Push(ImmediateVal(if (v) 1 else 0)))
      case NullLiter()               => List(Push(ImmediateVal(0)))
      case ArrayElem(value, indices) => List()
      case UnaryOp(x, op) =>
        op match {
          // TODO: chr and ord are TYPE CASTS. They do not change the internal value,
          //       but will need bound checking e.t.c.
          case UnaryOperator.Chr => List()
          case UnaryOperator.Ord => List()
          case UnaryOperator.Len => List()
          case UnaryOperator.Negate =>
            List(
              Negate(MemLocation(RSP, SizeDir.Word))
            )
          case UnaryOperator.Not =>
            evalExprOntoStack(x) ++
              List(
                Xor(MemLocation(RSP, SizeDir.Word), ImmediateVal(1))
              )

        }
      case BinaryOp(x, y, op) =>
        op match {
          case BinaryOperator.Add =>
            evalExprOntoStack(x) ++
              evalExprOntoStack(y) ++
              List(
                Pop(RAX),
                Add(MemLocation(RSP, SizeDir.Word), EAX)
                // TODO OVERFLOWING
              )
          case BinaryOperator.Sub =>
            evalExprOntoStack(x) ++
              evalExprOntoStack(y) ++
              List(
                Pop(RAX),
                Subtract(MemLocation(RSP, SizeDir.Word), EAX)
                // TODO OVERFLOWING
              )
          case BinaryOperator.Mul =>
            evalExprOntoStack(x) ++
              evalExprOntoStack(y) ++
              List(
                Pop(RAX),
                Multiply(EAX, MemLocation(RSP, SizeDir.Word)),
                Add(RSP, ImmediateVal(8)),
                Push(RAX)
                // TODO OVERFLOWING
              )
          case BinaryOperator.Div =>
            evalExprOntoStack(y) ++
              evalExprOntoStack(x) ++
              List(
                Pop(RAX),
                Divide(MemLocation(RSP, SizeDir.Word)),
                Add(RSP, ImmediateVal(8)),
                Push(RAX)
                // TODO CHECK DIVISOR IS NOT 0
              )
          case BinaryOperator.Mod =>
            evalExprOntoStack(y) ++
              evalExprOntoStack(x) ++
              List(
                Pop(RAX),
                Divide(MemLocation(RSP, SizeDir.Word)),
                Add(RSP, ImmediateVal(8)),
                Push(RDX)
                // TODO CHECK DIVISOR IS NOT 0
              )
          case BinaryOperator.Eq =>
            generateComparison(x, y, Cond.Equal)
          case BinaryOperator.Neq =>
            generateComparison(x, y, Cond.NotEqual)
          case BinaryOperator.Greater =>
            generateComparison(x, y, Cond.Greater)
          case BinaryOperator.GreaterEq =>
            generateComparison(x, y, Cond.GreaterEqual)
          case BinaryOperator.Less =>
            generateComparison(x, y, Cond.Less)
          case BinaryOperator.LessEq =>
            generateComparison(x, y, Cond.LessEqual)
          case BinaryOperator.And =>
            evalExprOntoStack(x) ++
              evalExprOntoStack(y) ++
              List(
                Pop(RAX),
                And(MemLocation(RSP, SizeDir.Word), EAX)
              )
          case BinaryOperator.Or =>
            evalExprOntoStack(x) ++
              evalExprOntoStack(y) ++
              List(
                Pop(RAX),
                Or(MemLocation(RSP, SizeDir.Word), EAX)
              )
        }
      case call: microWacc.Call => generateCall(call)
    }
  }

  def generateCall(call: microWacc.Call)(using
      stack: LinkedHashMap[Ident, Int],
      strings: ListBuffer[String]
  ): List[AsmLine] = {
    val argRegs = List(RDI, RSI, RDX, RCX, R8, R9)
    val microWacc.Call(target, args) = call
    argRegs.zip(args).flatMap { (reg, expr) =>
      evalExprOntoStack(expr) ++
        List(Pop(reg))
    } ++
      args.drop(argRegs.size).flatMap(evalExprOntoStack) ++
      List(assemblyIR.Call(LabelArg(labelGenerator.getLabel(target)))) ++
      (if (args.size > argRegs.size) {
         List(Load(RSP, IndexAddress(RSP, (args.size - argRegs.size) * 8)))
       } else Nil)
  }

  // def readIntoVar(dest: IndexAddress, readType: Builtin.ReadInt.type | Builtin.ReadChar.type)(using
  //     stack: LinkedHashMap[Ident, Int],
  //     strings: ListBuffer[String]
  // ): List[AsmLine] = {
  //   readType match {
  //     case Builtin.ReadInt =>
  //       strings += PrintFormat.Int.toString
  //     case Builtin.ReadChar =>
  //       strings += PrintFormat.Char.toString
  //   }
  //   List(
  //     Load(
  //       RDI,
  //       IndexAddress(
  //         RIP,
  //         LabelArg(s".L.str${strings.size - 1}")
  //       )
  //     ),
  //     Load(RSI, dest)
  //   ) ++
  //     // alignStack() ++
  //     List(assemblyIR.Call(CLibFunc.Scanf))

  // }

  def generateComparison(x: Expr, y: Expr, cond: Cond)(using
      stack: LinkedHashMap[Ident, Int],
      strings: ListBuffer[String]
  ): List[AsmLine] = {
    evalExprOntoStack(x) ++
      evalExprOntoStack(y) ++
      List(
        Pop(RAX),
        Compare(MemLocation(RSP, SizeDir.Word), EAX),
        Set(Register(RegSize.Byte, RegName.AL), cond),
        And(EAX, ImmediateVal(_8_BIT_MASK)),
        Push(RAX)
      )
  }
  def accessVar(ident: Ident)(using stack: LinkedHashMap[Ident, Int]): IndexAddress =
    IndexAddress(RSP, (stack.size - stack(ident)) * 8)

  def alignStack()(using stack: LinkedHashMap[Ident, Int]): List[AsmLine] = {
    List(
      And(RSP, ImmediateVal(-16))
    )
  }

  // Missing a sub instruction but dont think we need it
  def funcPrologue(): List[AsmLine] = {
    List(
      Push(RBP),
      Move(RBP, RSP)
    )
  }

  def funcEpilogue(): List[AsmLine] = {
    List(
      Move(RSP, RBP),
      Pop(RBP),
      assemblyIR.Return()
    )
  }

  // def saveRegs(regList: List[Register]): List[AsmLine] = regList.map(Push(_))
  // def restoreRegs(regList: List[Register]): List[AsmLine] = regList.reverse.map(Pop(_))

// TODO: refactor, really ugly function
//   def printF(expr: Expr)(using
//       stack: LinkedHashMap[Ident, Int],
//       strings: ListBuffer[String]
//   ): List[AsmLine] = {
// // determine the format string
//     expr.ty match {
//       case KnownType.String =>
//         strings += PrintFormat.String.toString
//       case KnownType.Char =>
//         strings += PrintFormat.Char.toString
//       case KnownType.Int =>
//         strings += PrintFormat.Int.toString
//       case _ =>
//         strings += PrintFormat.String.toString
//     }
//     List(
//       Load(
//         RDI,
//         IndexAddress(
//           RIP,
//           LabelArg(s".L.str${strings.size - 1}")
//         )
//       )
//     )
//       ++
//         // determine the actual value to print
//         (if (expr.ty == KnownType.Bool) {
//            expr match {
//              case BoolLiter(true) => {
//                strings += "true"
//              }
//              case _ => {
//                strings += "false"
//              }
//            }
//            List(
//              Load(
//                RDI,
//                IndexAddress(
//                  RIP,
//                  LabelArg(s".L.str${strings.size - 1}")
//                )
//              )
//            )

//          } else {
//            evalExprOntoStack(expr) ++
//              List(Pop(RSI))
//          })
//         // print the value
//         ++
//         List(
//           assemblyIR.Call(CLibFunc.PrintF),
//           Move(RDI, ImmediateVal(0)),
//           assemblyIR.Call(CLibFunc.Fflush)
//         )
//   }

// prints a new line
  // def printLn()(using
  //     stack: LinkedHashMap[Ident, Int],
  //     strings: ListBuffer[String]
  // ): List[AsmLine] = {
  //   strings += ""
  //   Load(
  //     RDI,
  //     IndexAddress(
  //       RIP,
  //       LabelArg(s".L.str${strings.size - 1}")
  //     )
  //   )
  //     ::
  //       List(
  //         assemblyIR.Call(CLibFunc.Puts),
  //         Move(RDI, ImmediateVal(0)),
  //         assemblyIR.Call(CLibFunc.Fflush)
  //       )

  // }
}
